/*
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "armboot.h"
#include "malloc.h"

typedef struct {
	ulong	size;
	env_t	env;
} board_env_t;

static board_env_t* get_env_ptr(void)
{
	static board_env_t* envbase = (board_env_t*) CFG_ENV_ADDR;
	
	while( envbase->size != 0 && envbase->size < CFG_ENV_SIZE && 
		envbase < (board_env_t*)(CFG_ENV_ADDR + CFG_ENV_SIZE))
	{
		env_t* env = &envbase->env;
		if( env->data[0] != 0 && env->crc == crc32(0,env->data,envbase->size-sizeof(ulong)) )
			return envbase;
			
		(uchar*)envbase += (envbase->size + sizeof(ulong));
	}
	envbase = (board_env_t*) CFG_ENV_ADDR;
	return NULL;
}

#if CFG_ENV_SIZE <= CFG_ENV_SECT_SIZE
#error Make sure that CFG_ENV_SIZE <= CFG_ENV_SECT_SIZE
#endif

/*
 * save environment buffer back to flash
 * returns -1 on error, 0 if ok
 */

#if (CONFIG_COMMANDS & CFG_CMD_SAVEENV)
static int get_env_length( env_t* env)
{
	int size = 0;

	if( env && env->data[0] != 0)
		while( size++ < CFG_ENV_SIZE )
			if( env->data[size++] == 0 && env->data[size++] == 0)
				return size;
	return 0;
}

static int save_env( env_t* env, int len, ulong addr)
{
	int rc = 0;
	
	if( len == 0 || (rc = flash_write((uchar*)&len, addr, sizeof(ulong)) ) < 0)
		return rc;

	addr += sizeof(ulong);
	
	return flash_write((uchar*)env, addr, len);
}

int board_env_save(bd_t *bd, env_t *env, int size)
{
	flash_info_t *finfo;
	ulong start_addr, end_addr;

	if( (finfo = addr2info(CFG_ENV_ADDR)) != NULL)
	{
		int rc = 0;
		ulong block = (finfo->size/finfo->sector_count);
		board_env_t* bd_env = get_env_ptr();
		
		// find start and end address of flash sector
		start_addr = finfo->start[CFG_ENV_ADDR/block];
		end_addr = start_addr + block - 1;	// last byte of sector

		// remove write protection
 		flash_sect_protect(0, start_addr, end_addr);
	
		printf("Saving Environment to Flash... ");

		// align to long, calculate crc
		size = (get_env_length(env) + 2*sizeof(ulong) -1) & ~3;
		env->crc = crc32(0,&env->data[0],size-sizeof(ulong));
		
		// try to write data to flash
		rc = save_env(env, size, bd_env? (ulong)bd_env : CFG_ENV_ADDR);

		if( bd_env && rc == ERR_NOT_ERASED)
		{
			ulong zero = 0;

			// make current env invalid
			flash_write( (uchar*)&zero, (ulong) &bd_env->env.data[0], sizeof(ulong));
			
			(uchar*)bd_env += (bd_env->size + sizeof(ulong));

			if( ((ulong)bd_env+size+sizeof(ulong)) < (CFG_ENV_ADDR+CFG_ENV_SIZE))
				rc = save_env(env, size, (ulong)bd_env);
		}
		
		if( rc == ERR_NOT_ERASED)
		{
			// try to erase flash keeping the rest
			uchar* pblock = (uchar*) malloc(block);

			flash_perror(rc);	// write flash not erased

			if( pblock != NULL)
			{
				memcpy( pblock, (uchar*)start_addr, block);

				if( (rc = flash_sect_erase(start_addr, end_addr)) > 0)
				{
					uchar* dest_addr = pblock + CFG_ENV_ADDR - start_addr;
					
					printf("Rewriting Flash... ");

					// copy environment data
					*((ulong*)dest_addr) = size;
					dest_addr += sizeof(ulong);
					
					memset( dest_addr, 0xff, CFG_ENV_SIZE);
					memcpy( dest_addr, (const uchar*)env, size);
					
					// try to write block to flash
					rc = flash_write( pblock, start_addr, block);
				}
				free(pblock);
			}
		}
		
		if (rc < 0)
			flash_perror(rc);
		else
			printf("done.\n");

		// protect environment sector
		flash_sect_protect(1, start_addr, end_addr);
	}
	return 0;
}
#endif

/*
 * copy environment to memory
 * returns -1 on error, 0 if ok
 */
int board_env_copy(bd_t *bd, env_t *data, int size)
{
    board_env_t *bd_env = get_env_ptr();

    if (bd_env != NULL)
    {
	memset(data, 0, sizeof(env_t));
	memcpy(data, &bd_env->env, bd_env->size);
	return 0;
    }
    return -1;
}

/*
 * try to read env character at offset #index
 *
 * called before the environment is copied to ram
 * returns -1 on error, 0 if ok
 */
int board_env_getchar(bd_t * bd, int index, uchar *c)
{
    board_env_t *bd_env = get_env_ptr();

    /* check environment crc */
    if (bd_env != NULL && index < (bd_env->size-sizeof(ulong)) )
    {
	env_t* env = &bd_env->env;
	*c = env->data[index];
	return 0;
    }
    return -1;
}

/*
 * try to read env character at offset #index
 *
 * called before the environment is copied to ram
 * returns -1 on error, 0 if ok
 */
uchar *board_env_getaddr(bd_t * bd, int index)
{
    board_env_t *bd_env = get_env_ptr();

    /* check environment crc */
    if (bd_env != NULL && index < (bd_env->size-sizeof(ulong)) )
    {
	env_t* env = &bd_env->env;
	return &env->data[index];
    }
    return 0;
}

