/*
 * (C) Copyright 2002
 * Kyle Harris, Nexus Technologies, Inc. kharris@nexus-tech.net
 *
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "armboot.h"

// handle two flash devs in parallel
#if CFG_FLASH_DEVS > 2
#error "configured to many flash devices!"
#endif

#if CFG_FLASH_DEVS > 1
#define FLASH_DEVS(cmd) ((cmd<<16)|cmd)
#else
#define FLASH_DEVS(cmd)	cmd
#endif

// flash device commands
#define FLASH_RESET_CMD		0x00FF
#define FLASH_SETUP_CMD		0x0040
#define FLASH_CONFIRM_CMD	0x00D0
#define FLASH_SUSPEND_CMD	0x00B0
#define FLASH_QUERY_CMD		0x0098
#define FLASH_STATUS_CMD	0x0070
#define FLASH_STATUS_READY	0x0080
#define FLASH_STATUS_CLEAR	0x0050
#define FLASH_ERASE_BLOCK	0x0020

#if CFG_MAX_FLASH_BANKS > 2
#error "configured to many flash banks!"
#endif

flash_info_t    flash_info[CFG_MAX_FLASH_BANKS];


/*-----------------------------------------------------------------------
 */

ulong flash_init(bd_t *bd)
{
    int i, j;
    ulong size = 0;

    for (i = 0; i < CFG_MAX_FLASH_BANKS; i++)
    {
	ulong secsize;
	ushort flashid,vendid;
	vulong *flashbase = 0;

	memset(flash_info[i].protect, 0, CFG_MAX_FLASH_SECT);

	switch (i)
        {
           case 0:
	        flashbase = (ulong*)PHYS_FLASH_1;
                break;
           case 1:
	        flashbase = (ulong*)PHYS_FLASH_2;
                break;
           default:
		return 0;
        }

	disable_interrupts();

	*flashbase = FLASH_DEVS(FLASH_QUERY_CMD);

#if CFG_FLASH_DEVS > 1
	vendid = (ushort) *flashbase;
	flashid = (ushort) *(flashbase+1);
	secsize = (*(flashbase+0x30) << 16);
#else
	vendid = (ushort) *flashbase;
	flashid = (ushort) (*flashbase >> 16);
	secsize = (*(flashbase+0x18) << 16);
#endif
	*flashbase = FLASH_DEVS(FLASH_RESET_CMD);

	enable_interrupts();

	flash_info[i].flash_id = ((vendid << 16) & FLASH_VENDMASK);
	flash_info[i].flash_id |= (flashid & FLASH_TYPEMASK);
	flash_info[i].size = CFG_FLASH_DEVS * (0x400000 << (flashid-0x16));

	secsize *= CFG_FLASH_DEVS;
	flash_info[i].sector_count = flash_info[i].size / secsize;

	for (j = 0; j < flash_info[i].sector_count; j++)
	    flash_info[i].start[j] = (ulong)flashbase + j * secsize;

	size += flash_info[i].size;
    }

    /* Protect monitor and environment sectors
     */
    flash_protect(FLAG_PROTECT_SET,
		  CFG_FLASH_BASE,
		  CFG_FLASH_BASE + _armboot_end - _armboot_start,
		  &flash_info[0]);

    flash_protect(FLAG_PROTECT_SET,
		  CFG_ENV_ADDR,
		  CFG_ENV_ADDR + CFG_ENV_SIZE - 1,
		  &flash_info[0]);

    return size;
}

/*-----------------------------------------------------------------------
 */
void flash_print_info  (flash_info_t *info)
{
    int i, j;

    for (j=0; j<CFG_MAX_FLASH_BANKS; j++)
    {
        switch (info->flash_id & FLASH_VENDMASK)
        {
        case (INTEL_MANUFACT & FLASH_VENDMASK):
	        printf("Intel: ");
	        break;
        default:
	        printf("Unknown Vendor ");
	        break;
        }

        switch (info->flash_id & FLASH_TYPEMASK)
        {
	case (INTEL_ID_28F320JA3 & FLASH_TYPEMASK):
		printf("%d x 28F320J3 (32Mbit)\n",CFG_FLASH_DEVS); break;
	case (INTEL_ID_28F640JA3 & FLASH_TYPEMASK):
		printf("%d x 28F640J3 (64Mbit)\n",CFG_FLASH_DEVS); break;
        case (INTEL_ID_28F128JA3 & FLASH_TYPEMASK):
	        printf("%d x 28F128J3 (128Mbit)\n",CFG_FLASH_DEVS); break;
        default:
	        printf("Unknown Chip Type\n");
	        return;
        }

        printf("  Size: %ld MB in %d Sectors\n",
	        info->size >> 20, info->sector_count);

        printf("  Sector Start Addresses:");
        for (i = 0; i < info->sector_count; i++)
        {
	        if ((i % 5) == 0)
	        	printf ("\n   ");

	        printf (" %08lX%s", info->start[i],
		        info->protect[i] ? " (RO)" : "     ");
        }
        printf ("\n");
        info++;
    }
}

/*-----------------------------------------------------------------------
 */

int	flash_erase (flash_info_t *info, int s_first, int s_last)
{
    int flag, prot, sect;
    int rc = ERR_OK;

    if (info->flash_id == FLASH_UNKNOWN)
	return ERR_UNKNOWN_FLASH_TYPE;

    if ((s_first < 0) || (s_first > s_last)) {
	return ERR_INVAL;
    }

    if ((info->flash_id & FLASH_VENDMASK) !=
	(INTEL_MANUFACT & FLASH_VENDMASK)) {
	return ERR_UNKNOWN_FLASH_VENDOR;
    }

    prot = 0;
    for (sect=s_first; sect<=s_last; ++sect) {
	if (info->protect[sect]) {
	    prot++;
	}
    }
    if (prot)
	return ERR_PROTECTED;

    /*
     * Disable interrupts which might cause a timeout
     * here. Remember that our exception vectors are
     * at address 0 in the flash, and we don't want a
     * (ticker) exception to happen while the flash
     * chip is in programming mode.
     */
    flag = disable_interrupts();

    /* Start erase on unprotected sectors */
    for (sect = s_first; sect<=s_last && !ctrlc(); sect++) {

	printf("Erasing sector %2d ... ", sect);

	/* arm simple, non interrupt dependent timer */
	reset_timer_masked();

	if (info->protect[sect] == 0) {	/* not protected */
	    vulong *addr = (vulong *)(info->start[sect]);

	    *addr = FLASH_DEVS(FLASH_ERASE_BLOCK);	/* erase setup */
	    *addr = FLASH_DEVS(FLASH_CONFIRM_CMD);	/* erase confirm */

	    while ((*addr & FLASH_DEVS(FLASH_STATUS_READY)) != FLASH_DEVS(FLASH_STATUS_READY)) {
		if (get_timer_masked() > CFG_FLASH_ERASE_TOUT) {
		    *addr = FLASH_DEVS(FLASH_SUSPEND_CMD); /* suspend erase */
		    *addr = FLASH_DEVS(FLASH_RESET_CMD);	/* reset to read mode */
		    rc = ERR_TIMOUT;
		    goto outahere;
		}
	    }

	    /* clear status register command */
	    *addr = FLASH_DEVS(FLASH_STATUS_CLEAR);
	    /* reset to read mode */
	    *addr = FLASH_DEVS(FLASH_RESET_CMD);
	}
	printf("ok.\n");
    }
    if (ctrlc())
      printf("User Interrupt!\n");

outahere:

    /* allow flash to settle - wait 10 ms */
    udelay_masked(10000);

    if (flag)
      enable_interrupts();

    return rc;
}

/*-----------------------------------------------------------------------
 * Copy memory to flash
 */

#if CFG_FLASH_DEVS > 1

static int write_long (flash_info_t *info, ulong dest, ulong data)
{
    vulong *addr = (vulong *)dest;
    ulong barf;
    int rc = ERR_OK;
    int flag;

    /* Check if Flash is (sufficiently) erased
     */
    if ((*addr & data) != data)
        return ERR_NOT_ERASED;

    /*
     * Disable interrupts which might cause a timeout
     * here. Remember that our exception vectors are
     * at address 0 in the flash, and we don't want a
     * (ticker) exception to happen while the flash
     * chip is in programming mode.
     */
    flag = disable_interrupts();

    /* clear status register command */
    *addr = FLASH_DEVS(FLASH_STATUS_CLEAR);

    /* program set-up command */
    *addr = FLASH_DEVS(FLASH_SETUP_CMD);

    /* latch address/data */
    *addr = data;

    /* arm simple, non interrupt dependent timer */
    reset_timer_masked();

    /* wait while polling the status register */
    while(((barf = *addr) & FLASH_DEVS(FLASH_STATUS_READY)) != FLASH_DEVS(FLASH_STATUS_READY))
    {
	if (get_timer_masked() > CFG_FLASH_WRITE_TOUT) {
	    rc = ERR_TIMOUT;
	    /* suspend program command */
	    *addr = FLASH_DEVS(FLASH_SUSPEND_CMD);
	    goto outahere;
	}
    }
	if( barf & FLASH_DEVS(0x001A)) {	/* check for error */
	    if( barf & 0x001A0000) {
		barf >>=16;
	    } else {
		barf &= 0x0000001A;
	    }
	    printf("\nFlash write error %02lx at address %08lx\n",
		   barf, (unsigned long)dest);
	    if(barf & 0x0002) {
		printf("Device protect error.\n");
		rc = ERR_NOT_ERASED;
		goto outahere;
	    }
	    if(barf & 0x0010) {
		printf("Programming error.\n");
		rc = ERR_PROG_ERROR;
		goto outahere;
	    }
	    if(barf & 0x0008) {
		printf("Voltage range error.\n");
		rc = ERR_PROG_ERROR;
		goto outahere;
	    }
	    rc = ERR_PROG_ERROR;
	    goto outahere;
	}

outahere:
    /* read array command */
    *addr = FLASH_DEVS(FLASH_RESET_CMD);

    if (flag)
      enable_interrupts();

    return rc;
}

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
    ulong cp, wp, data;
    int l;
    int i, rc;

    wp = (addr & ~3);	/* get lower word aligned address */

    /*
     * handle unaligned start bytes
     */
    if ((l = addr - wp) != 0) {
	data = 0;
	for (i=0, cp=wp; i<l; ++i, ++cp) {
	    data = (data >> 8) | (*(uchar *)cp << 24);
	}
	for (; i<4 && cnt>0; ++i) {
	    data = (data >> 8) | (*src++ << 24);
	    --cnt;
	    ++cp;
	}
	for (; cnt==0 && i<4; ++i, ++cp) {
	    data = (data >> 8) | (*(uchar *)cp << 24);
	}

	if ((rc = write_long(info, wp, data)) != 0) {
	    return (rc);
	}
	wp += 4;
    }

    /*
     * handle word aligned part
     */
    while (cnt >= 4) {
	data = *((vulong*)src);
	if ((rc = write_long(info, wp, data)) != 0) {
	    return (rc);
	}
	src += 4;
	wp  += 4;
	cnt -= 4;
    }

    if (cnt == 0) {
	return ERR_OK;
    }

    /*
     * handle unaligned tail bytes
     */
    data = 0;
    for (i=0, cp=wp; i<4 && cnt>0; ++i, ++cp) {
	data = (data >> 8) | (*src++ << 24);
	--cnt;
    }
    for (; i<4; ++i, ++cp) {
	data = (data >> 8) | (*(uchar *)cp << 24);
    }

    return write_long(info, wp, data);
}

#else

static int write_word (flash_info_t *info, ulong dest, ushort data)
{
    vushort *addr = (vushort *)dest, val;
    int rc = ERR_OK;
    int flag;

    /* Check if Flash is (sufficiently) erased
     */
    if ((*addr & data) != data)
        return ERR_NOT_ERASED;

    /*
     * Disable interrupts which might cause a timeout
     * here. Remember that our exception vectors are
     * at address 0 in the flash, and we don't want a
     * (ticker) exception to happen while the flash
     * chip is in programming mode.
     */
    flag = disable_interrupts();

    /* clear status register command */
    *addr = FLASH_STATUS_CLEAR;

    /* program set-up command */
    *addr = FLASH_SETUP_CMD;

    /* latch address/data */
    *addr = data;

    /* arm simple, non interrupt dependent timer */
    reset_timer_masked();

    /* wait while polling the status register */
    while(((val = *addr) & FLASH_STATUS_READY) != FLASH_STATUS_READY)
    {
	if (get_timer_masked() > CFG_FLASH_WRITE_TOUT) {
	    rc = ERR_TIMOUT;
	    /* suspend program command */
	    *addr = FLASH_SUSPEND_CMD;
	    goto outahere;
	}
    }

    if(val & 0x1A) {	/* check for error */
        printf("\nFlash write error %02x at address %08lx\n",
    	   (int)val, (unsigned long)dest);
        if(val & (1<<3)) {
	    printf("Voltage range error.\n");
	    rc = ERR_PROG_ERROR;
	    goto outahere;
        }
        if(val & (1<<1)) {
	    printf("Device protect error.\n");
	    rc = ERR_PROTECTED;
	    goto outahere;
        }
        if(val & (1<<4)) {
	    printf("Programming error.\n");
	    rc = ERR_PROG_ERROR;
	    goto outahere;
        }
        rc = ERR_PROG_ERROR;
        goto outahere;
    }

outahere:
    /* reset array command */
    *addr = FLASH_RESET_CMD;

    if (flag)
      enable_interrupts();

    return rc;
}

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
    ulong cp, wp;
    ushort data;
    int l;
    int i, rc;

    wp = (addr & ~1);	/* get lower word aligned address */

    /*
     * handle unaligned start bytes
     */
    if ((l = addr - wp) != 0) {
	data = 0;
	for (i=0, cp=wp; i<l; ++i, ++cp) {
	    data = (data >> 8) | (*(uchar *)cp << 8);
	}
	for (; i<2 && cnt>0; ++i) {
	    data = (data >> 8) | (*src++ << 8);
	    --cnt;
	    ++cp;
	}
	for (; cnt==0 && i<2; ++i, ++cp) {
	    data = (data >> 8) | (*(uchar *)cp << 8);
	}

	if ((rc = write_word(info, wp, data)) != 0) {
	    return (rc);
	}
	wp += 2;
    }

    /*
     * handle word aligned part
     */
    while (cnt >= 2) {
	data = *((vushort*)src);
	if ((rc = write_word(info, wp, data)) != 0) {
	    return (rc);
	}
	src += 2;
	wp  += 2;
	cnt -= 2;
    }

    if (cnt == 0) {
	return ERR_OK;
    }

    /*
     * handle unaligned tail bytes
     */
    data = 0;
    for (i=0, cp=wp; i<2 && cnt>0; ++i, ++cp) {
	data = (data >> 8) | (*src++ << 8);
	--cnt;
    }
    for (; i<2; ++i, ++cp) {
	data = (data >> 8) | (*(uchar *)cp << 8);
    }

    return write_word(info, wp, data);
}
#endif
