// C2 interface implementation to program a C8051F30x device

#include "armboot.h"
#include "pxa-regs.h"

#define C2D		(GPLR0 & GPIO_bit(25))
#define C2D_LOW		GPCR0 = GPIO_bit(25)
#define C2D_HIGH	GPSR0 = GPIO_bit(25)
#define C2D_DRIVER_ON	GPDR0 |= GPIO_bit(25)
#define C2D_DRIVER_OFF	GPDR0 &= ~GPIO_bit(25)

#define C2CK_LOW	GPCR0 = GPIO_bit(23)
#define C2CK_HIGH	GPSR0 = GPIO_bit(23)
#define C2CK_STROBE	C2CK_LOW; C2CK_HIGH;
#define C2CK_DRIVER_ON  GPDR0 |= GPIO_bit(23)

// C2 status return codes
#define  C2_CMD_INVALID   0x00
#define  C2_CMD_FAILED    0x02
#define  C2_CMD_OK        0x0D

// C2 interface commands
#define  C2_GET_VERSION       0x01
#define  C2_BLOCK_READ        0x06
#define  C2_BLOCK_WRITE       0x07
#define  C2_PAGE_ERASE        0x08
#define  C2_DEVICE_ERASE      0x03

// C2 Registers
#define  C2_FPDAT             0xB4
#define  C2_FPCTL             0x02
#define  C2_DEVICEID          0x00
#define  C2_REVID             0x01

//-----------------------------------------------------------------------------------
// c2_Reset() - performs a target device reset by pulling the C2CK pin low for >20us

static void c2_Reset(void)
{
	C2CK_DRIVER_ON;
	C2D_DRIVER_OFF;
	udelay(1);
	
	C2CK_LOW;	// put target device in reset state by pulling
	udelay(50);	// C2CK low for >20us
	C2CK_HIGH;	// release target device from reset
	udelay(5);	// wait for >2us
}

//-----------------------------------------------------------------------------------
// c2_ReadAR() - performs a C2 Address register read, returns the 8-bit register content

static unsigned char c2_ReadAR(void)
{
	unsigned char i;	// bit counter
	unsigned char addr;	// address register read content

	// START field
	C2CK_STROBE;		// strobe C2CK with C2D driver disabled

	// INS field (10b, LSB first)
	C2D_LOW;
	C2D_DRIVER_ON;	// enable C2D driver (output)
	C2CK_STROBE;
	C2D_HIGH;
	C2CK_STROBE;
	C2D_DRIVER_OFF;		// disable C2D driver (input)

	// ADDRESS field
	addr = 0;
	for( i=0; i<8; i++)	// shift in 8 bit ADDRESS field, LSB first
	{              
		addr >>= 1;
		C2CK_STROBE;
		if(C2D > 0)
			addr |= 0x80;
	}

	// STOP field
	C2CK_STROBE;

	return addr;
}

//-----------------------------------------------------------------------------------
// c2_WriteAR() - performs a C2 Address register write (writes the <addr> input to Address register)

static void c2_WriteAR(unsigned char addr)
{
	unsigned char i;                    // bit counter

	// START field
	C2CK_STROBE;                         // strobe C2CK with C2D driver disabled

	// INS field (11b, LSB first)
	C2D_HIGH;
	C2D_DRIVER_ON;	// enable C2D driver (output)
	C2CK_STROBE;
	C2D_HIGH;
	C2CK_STROBE;

	// ADDRESS field
	for( i=0; i<8; i++)	// shift out 8-bit ADDRESS field
	{
		if(addr & 0x01)
			C2D_HIGH;
		else
			C2D_LOW;
		C2CK_STROBE;
		addr >>= 1;
	}

	// STOP field
	C2D_DRIVER_OFF;         // disable C2D driver (input)
	C2CK_STROBE;
}

//-----------------------------------------------------------------------------------
// c2_ReadDR() - performs a C2 Data register read, returns the 8-bit register content

static unsigned char c2_ReadDR(void)
{
	unsigned char i;	// bit counter
	unsigned char dat;	// data register read content

	// START field
	C2CK_STROBE;                         // strobe C2CK with C2D driver disabled

	// INS field (00b, LSB first)
	C2D_LOW;
	C2D_DRIVER_ON;	// enable C2D driver (output)
	C2CK_STROBE;
	C2D_LOW;
	C2CK_STROBE;

	// LENGTH field (00b -> 1 byte)
	C2D_LOW;
	C2CK_STROBE;
	C2D_LOW;
	C2CK_STROBE;

	// WAIT field
	C2D_DRIVER_OFF;		// disable C2D driver for input
	do
	{
		C2CK_STROBE;
	} while(C2D == 0);	// strobe C2CK until target transmits a '1'

	// DATA field
	dat = 0;
	for( i=0; i<8; i++)	// shift in 8 bit DATA field, LSB first
	{              
		dat >>= 1;
		C2CK_STROBE;
		if(C2D > 0)
			dat |= 0x80;
	}

	// STOP field
	C2CK_STROBE;

	return dat;
}

//-----------------------------------------------------------------------------------
// c2_WriteDR() - performs a C2 Data register write (writes <dat> input to data register)

static void c2_WriteDR(unsigned char dat)
{
	unsigned char i;                    // Bit counter

	// START field
	C2CK_STROBE;                         // strobe C2CK with C2D driver disabled

	// INS field (01b, LSB first)
	C2D_HIGH;
	C2D_DRIVER_ON;	// enable C2D driver (output)
	C2CK_STROBE;
	C2D_LOW;
	C2CK_STROBE;

	// LENGTH field (00b -> 1 byte)
	C2D_LOW;
	C2CK_STROBE;
	C2D_LOW;
	C2CK_STROBE;

	// DATA field
	for( i=0; i<8; i++)	// shift out 8-bit ADDRESS field
	{
		if(dat & 0x01)
			C2D_HIGH;
		else
			C2D_LOW;
		C2CK_STROBE;
		dat >>= 1;
	}

	// WAIT field
	C2D_DRIVER_OFF;		// disable C2D driver for input
	do
	{
		C2CK_STROBE;
	} while(C2D == 0);	// strobe C2CK until target transmits a '1'

	// STOP field
	C2CK_STROBE;
}

//-----------------------------------------------------------------------------------
// c2_GetDevID() - reads the target Devcie ID register

unsigned char c2_GetDevID(void)
{
	c2_Reset();
	c2_WriteAR(C2_DEVICEID);	// select device ID regsiter for C2 Data register accesses
	return c2_ReadDR();	// read and return the device ID register
}

//-----------------------------------------------------------------------------------
// c2_GetRevID() - reads the target Revision ID register

unsigned char c2_GetRevID(void)
{
	c2_Reset();
	c2_WriteAR(C2_REVID);	// select device ID regsiter for C2 Data register accesses
	return c2_ReadDR();	// read and return the device ID register
}

//-----------------------------------------------------------------------------------
// c2_FlashInit() - initializes the C2 interface for FLASH programming

static void c2_FlashInit(void)
{
	c2_Reset();		// reset the target device
	c2_WriteAR(C2_FPCTL);	// target the C2 FLASH Programming
   
	// enable C2 FLASH programing
	c2_WriteDR(0x02);
	c2_WriteDR(0x01);

	udelay(30000);	// wait for > 20ms
}

#define	Poll_OutReady	while(!(c2_ReadAR()&0x01))
#define	Poll_InBusy	while((c2_ReadAR()&0x02))

//-----------------------------------------------------------------------------------
// c2_Device_Erase() - erases the entire FLASH memory space

int c2_DeviceErase(void)
{
	c2_FlashInit();
	
	c2_WriteAR(C2_FPDAT);		// select the FLASH programming data register 
	c2_WriteDR(C2_DEVICE_ERASE);	// send device erase command
	Poll_InBusy;			// wait for input acknowledge

	// check status before starting FLASH access sequence
	Poll_OutReady;			// wait for status information
	if( c2_ReadDR() != C2_CMD_OK)	// read FLASH programming interface status
		return 0;

	// send a three-byte arming sequence (0xDE, 0xAD, 0xA5) to enable the device erase
	c2_WriteDR(0xDE);
	Poll_InBusy;

	c2_WriteDR(0xAD);
	Poll_InBusy;

	c2_WriteDR(0xA5);
	Poll_InBusy;

	Poll_OutReady;
	return 1;
}

//-----------------------------------------------------------------------------------
// c2_BlockRead() - reads a block of FLASH memory

int c2_BlockRead(unsigned int flash_addr, unsigned char* buffer, int length)
{
	int i;

	c2_FlashInit();
	
	c2_WriteAR(C2_FPDAT);		// select the FLASH programming data register 
	c2_WriteDR(C2_BLOCK_READ);	// send FLASH block read command
	Poll_InBusy;			// wait for input acknowledge

	// check status before starting FLASH access sequence
	Poll_OutReady;			// wait for status information
	if( c2_ReadDR() != C2_CMD_OK)	// read FLASH programming interface status
		return 0;

	c2_WriteDR(flash_addr >> 8);		// send address high byte to FPDAT
	Poll_InBusy;				// wait for input acknowledge
	c2_WriteDR(flash_addr & 0xff);		// send address low byte to FPDAT
	Poll_InBusy;				// wait for input acknowledge
	c2_WriteDR(length);			// send block size
	Poll_InBusy;				// wait for input acknowledge

	// check status before reading FLASH block
	Poll_OutReady;			// wait for status information
	if( c2_ReadDR() != C2_CMD_OK)	// read FLASH programming interface status
		return 0;

	// read FLASH block
	for( i=0; i<length; i++)
	{
		Poll_OutReady;			// wait for data ready indicator
		*buffer++ = c2_ReadDR();	// read data from the FPDAT register
	}
	
	return length;		// exit and indicate success
}

//-----------------------------------------------------------------------------------
// c2_BlockWrite() - writes a block of FLASH memory

int c2_BlockWrite(unsigned int flash_addr, unsigned char* buffer, int length)
{
	int i;

	c2_FlashInit();

	c2_WriteAR(C2_FPDAT);		// select the FLASH programming data register 
	c2_WriteDR(C2_BLOCK_WRITE);	// send FLASH block write command
	Poll_InBusy;			// wait for input acknowledge

	// check status before starting FLASH access sequence
	Poll_OutReady;			// wait for status information
	if( c2_ReadDR() != C2_CMD_OK)	// read FLASH programming interface status
		return 0;

	c2_WriteDR(flash_addr >> 8);		// send address high byte to FPDAT
	Poll_InBusy;				// wait for input acknowledge
	c2_WriteDR(flash_addr & 0x00FF);	// send address low byte to FPDAT
	Poll_InBusy;				// wait for input acknowledge
	c2_WriteDR(length);			// send block size
	Poll_InBusy;				// wait for input acknowledge

	// check status before reading FLASH block
	Poll_OutReady;			// wait for status information
	if( c2_ReadDR() != C2_CMD_OK)	// read FLASH programming interface status
		return 0;

	// write FLASH block
	for( i=0; i<length; i++)
	{
		c2_WriteDR(*buffer++);		// write data to the FPDAT register
		Poll_InBusy;			// wait for input acknowledge
	}

	Poll_OutReady;		// wait for last FLASH write to complete
	return length;		// exit and indicate success
}

