/*
 * Very simple (yet, for some reason, very effective) memory tester.
 * Originally by Simon Kirby <sim@stormix.com> <sim@neato.org>
 * Version 2 by Charles Cazabon <memtest@discworld.dyndns.org>
 *
 * This file contains the functions for the actual tests, called from the
 * main routine in memtest.c.  See other comments in that file.
 *
 */


/* Includes. */
#include <armboot.h>
#include "memtest.h"


/* Data. */
int silent_test = 0;

/* Data. */
teststruct_0 tests_noarg[] =
{
	{ "Stuck Address", test_stuck_address },
	{ "Random value", test_random_value },
	{ "XOR comparison", test_xor_comparison },
	{ "SUB comparison", test_sub_comparison },
	{ "MUL comparison", test_mul_comparison },
	{ "DIV comparison",test_div_comparison },
	{ "OR comparison", test_or_comparison },
	{ "AND comparison", test_and_comparison },
	{ "Sequential Increment", test_seqinc_comparison },
	{ "Solid Bits", test_solidbits_comparison },
	{ "Block Sequential", test_blockseq_comparison },
	{ "Checkerboard", test_checkerboard_comparison },
	{ "Bit Spread", test_bitspread_comparison },
	{ "Bit Flip", test_bitflip_comparison },
	{ NULL, 0 }
};

teststruct_1 tests_1arg[] =
{
	{ "Walking Ones",test_walkbits_comparison, 0 },
	{ "Walking Zeroes",test_walkbits_comparison, 1 },
	{ NULL, 0, 0 }
};

#define RNG_M	2147483647L
#define RNG_A	48271L
#define RNG_Q	127773L
#define RNG_R	2836L

static int rnd_seed;

int rand(void)
{
	int low, high, test;

	high = rnd_seed / RNG_Q;
	low = rnd_seed % RNG_Q;
	test = RNG_A * low - RNG_R * high;

	return rnd_seed = test > 0 ? test : test + RNG_M;
}

/* Function definitions. */
int
test_verify_success (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	
	if (!silent_test) printf ("\b\b\b\b\b\b\b\b\b\bTesting...");

	for (i = 0; i < count; i++, p1++, p2++)
	{
		if (*p1 != *p2)
		{
			printf ( "\nFAILURE: 0x%08lx != 0x%08lx at offset 0x%08lx.\n", *p1, *p2, i);
			printf ( "Skipping to next test...\n");
			return (ERROR);
		}
	}
	return (OK);
}


int
test_random_value (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	
	silent_test = 0;		
	for (i = 0; i < count; i++)
	{
		*p1++ = *p2++ = RAND32;
	}
	return (test_verify_success (bp1, bp2, count));
}


int
test_xor_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	ui32 q = RAND32;
	
	silent_test = 0;		
	for (i = 0; i < count; i++)
	{
		*p1++ ^= q;
		*p2++ ^= q;
	}
	return (test_verify_success (bp1, bp2, count));
}


int
test_sub_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	ui32 q = RAND32;
	
	silent_test = 0;		
	for (i = 0; i < count; i++)
	{
		*p1++ -= q;
		*p2++ -= q;
	}
	return (test_verify_success (bp1, bp2, count));
}


int
test_mul_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	ui32 q = RAND32;
	
	silent_test = 0;		
	for (i = 0; i < count; i++)
	{
		*p1++ *= q;
		*p2++ *= q;
	}
	return (test_verify_success (bp1, bp2, count));
}


int
test_div_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	ui32 q = RAND32;
	
	silent_test = 0;		
	for (i = 0; i < count; i++)
	{
		if (!q)
			q++;
		*p1++ /= q;
		*p2++ /= q;
	}
	return (test_verify_success (bp1, bp2, count));
}


int
test_or_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	ui32 q = RAND32;
	
	silent_test = 0;		
	for (i = 0; i < count; i++)
	{
		*p1++ |= q;
		*p2++ |= q;
	}
	return (test_verify_success (bp1, bp2, count));
}


int
test_and_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	ui32 q = RAND32;
	
	silent_test = 0;		
	for (i = 0; i < count; i++)
	{
		*p1++ &= q;
		*p2++ &= q;
	}
	return (test_verify_success (bp1, bp2, count));
}


int
test_seqinc_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i;
	ui32 q = RAND32;
	
	silent_test = 0;		
	for (i = 0; i < count; i++)
	{
		*p1++ = *p2++ = (i + q);
	}
	return (test_verify_success (bp1, bp2, count));
}


int
test_solidbits_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 q, i, j;

	silent_test = 1;		
	printf ("    ");
	for (j = 0; j < 64; j++)
	{
		q = (j % 2) == 0 ? 0xFFFFFFFF : 0x00000000;
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bSetting... %3lu", j);
		p1 = (volatile ui32 *) bp1;
		p2 = (volatile ui32 *) bp2;
		for (i = 0; i < count; i++)
		{
			*p1++ = *p2++ = (i % 2) == 0 ? q : ~q;
		}
	
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bTesting... %3lu", j);

		if (test_verify_success (bp1, bp2, count) == ERROR)
		{
			return (ERROR);
		}
	}
	printf ( "\b\b\b\b");
	return (OK);
}


int
test_checkerboard_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 q, i, j;

	silent_test = 1;		
	printf ( "    ");
	for (j = 0; j < 64; j++)
	{
		q = (j % 2) == 0 ? 0x55555555 : 0xAAAAAAAA;
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bSetting... %3lu", j);
		p1 = (volatile ui32 *) bp1;
		p2 = (volatile ui32 *) bp2;
		for (i = 0; i < count; i++)
		{
			*p1++ = *p2++ = (i % 2) == 0 ? q : ~q;
		}
	
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bTesting... %3lu", j);

		if (test_verify_success (bp1, bp2, count) == ERROR)
		{
			return (ERROR);
		}
	}
	printf ( "\b\b\b\b");
	return (OK);
}


int
test_blockseq_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i, j;
	
	silent_test = 1;		
	printf ( "    ");
	for (j = 0; j < 256; j++)
	{
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bSetting... %3lu", j);
		p1 = (volatile ui32 *) bp1;
		p2 = (volatile ui32 *) bp2;
		for (i = 0; i < count; i++)
		{
			*p1++ = *p2++ = MAKE32FROM8 (j);
		}
	
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bTesting... %3lu", j);

		if (test_verify_success (bp1, bp2, count) == ERROR)
		{
			return (ERROR);
		}
	}
	printf ( "\b\b\b\b");
	return (OK);
}


int
test_walkbits_comparison (ui32 *bp1, ui32 *bp2, ui32 count, int m)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i, j;

	silent_test = 1;		
	printf ( "    ");
	for (j = 0; j < 64; j++)
	{
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bSetting... %3lu", j);
		p1 = (volatile ui32 *) bp1;
		p2 = (volatile ui32 *) bp2;
		for (i = 0; i < count; i++)
		{
			if (j < 32)	/* Walk it up. */
			{
				*p1++ = *p2++ = (m == 0) ? 0x00000001 << j :
										   0xFFFFFFFF ^ (0x00000001 << j);
			}
			else		/* Walk it back down. */
			{
				*p1++ = *p2++ = (m == 0)
					? 0x00000001 << (64 - j - 1)
					: 0xFFFFFFFF ^ (0x00000001 << (64 - j - 1));
			}				
		}

		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bTesting... %3lu", j);
		if (test_verify_success (bp1, bp2, count) == ERROR)
		{
			return (ERROR);
		}
	}
	printf ( "\b\b\b\b");
	return (OK);
}


int
test_bitspread_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i, j;

	silent_test = 1;		
	printf ( "    ");
	for (j = 0; j < 64; j++)
	{
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bSetting... %3lu", j);
		p1 = (volatile ui32 *) bp1;
		p2 = (volatile ui32 *) bp2;
		for (i = 0; i < count; i++)
		{
			if (j < 32)	/* Walk it up. */
			{
				*p1++ = *p2++ = (i % 2 == 0)
					? (0x00000001 << j) | (0x00000001 << (j + 2))
					: 0xFFFFFFFF ^ ((0x00000001 << j)
									| (0x00000001 << (j + 2)));
			}
			else		/* Walk it back down. */
			{
				*p1++ = *p2++ = (i % 2 == 0)
					? (0x00000001 << (63 - j)) | (0x00000001 << (65 - j))
					: 0xFFFFFFFF ^ (0x00000001 << (63 - j) 
									| (0x00000001 << (65 - j)));
			}				
		}

		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bTesting... %3lu", j);
		if (test_verify_success (bp1, bp2, count) == ERROR)
		{
			return (ERROR);
		}
	}
	printf ( "\b\b\b\b");
	return (OK);
}


int
test_bitflip_comparison (ui32 *bp1, ui32 *bp2, ui32 count)
{
	volatile ui32 *p1 = (volatile ui32 *) bp1;
	volatile ui32 *p2 = (volatile ui32 *) bp2;
	ui32 i, j, k;
	ui32 q;
		
	silent_test = 1;		
	printf ( "    ");
	for (k = 0; k < 32; k++)
	{
		q = 0x00000001 << k;

		for (j = 0; j < 8; j++)
		{
			q = ~q;
			printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bSetting... %3lu",
					 k * 8 + j);
			p1 = (volatile ui32 *) bp1;
			p2 = (volatile ui32 *) bp2;
			for (i = 0; i < count; i++)
			{
				*p1++ = *p2++ = (i % 2) == 0 ? q : ~q;
			}
		
			printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bTesting... %3lu",
					 k * 8 + j);
	
			if (test_verify_success (bp1, bp2, count) == ERROR)
			{
				return (ERROR);
			}
		}
		
	}
	printf ( "\b\b\b\b");
	return (OK);
}


int
test_stuck_address (ui32 *bp1, ui32 *unused, ui32 count)
{
	volatile ui32 *p1;
	/* second argument is not used; just gives it a compatible signature. */
	ui32 i, j;

	count <<= 1;
	printf ( "    ");
	for (j = 0; j < 16; j++)
	{
		p1 = (volatile ui32 *) bp1;
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bSetting... %3lu", j);
		for (i = 0; i < count; i++)
		{
			*p1++ = ((j + i) % 2) == 0 ? (ui32) p1 : ~((ui32) p1);
		}
		printf ( "\b\b\b\b\b\b\b\b\b\b\b\b\b\bTesting... %3lu", j);
		p1 = (volatile ui32 *) bp1;
		for (i = 0; i < count; i++, p1++)
		{
			if (*p1 != (((j + i) % 2) == 0 ? (ui32) p1 : ~((ui32) p1)))
			{
				printf ("\nFAILURE: possible bad address line at offset 0x%08lx.\n", i);
				printf ("Skipping to next test...\n");
				return (ERROR);
			}
		}
	}
	printf ( "\b\b\b\b");
	return (OK);
}
