
#include "armboot.h"
#include "command.h"
#include "ax88796.h"
#include "net.h"

#ifdef CONFIG_DRIVER_AX88796

#if (CONFIG_COMMANDS & CFG_CMD_NET)

#define NE_DATAPORT     EI_SHIFT(0x10)  /* NatSemi-defined port window offset. */
#define NE_RESET        EI_SHIFT(0x1f)  /* Issue a read to reset, a write to clear. */

static unsigned char inb( unsigned addr)
{
	return *((volatile unsigned short *) addr);
}

static void insw( unsigned addr, unsigned short* buffer, int count)
{
	while( count-- > 0)
		*buffer++ = *((volatile unsigned short *)addr);
}

static void outb( unsigned char c, unsigned addr)
{
	*((volatile unsigned short *)addr) = c;
}

static void outsw( unsigned addr, unsigned short* buffer, int count)
{
	while( count-- > 0)
		*((volatile unsigned short *)addr) = *buffer++;
}

void eth_halt(void)
{
#ifdef DEBUG
	printf( "ax88796: eth_halt\n");
#endif
	outb(E8390_NODMA+E8390_PAGE0+E8390_STOP, E8390_BASE + E8390_CMD);       // switch to page 0 + stop
}

static int ax88796_reset(unsigned ioaddr)
{
	unsigned rto = get_timer(0) + 2*CFG_HZ/100;		// reset timeout 20ms
	
	// reset card, who knows what dain-bramaged state it was left in
	outb(inb(ioaddr + NE_RESET), ioaddr + NE_RESET);
	
	while ((inb(ioaddr + EN0_ISR) & ENISR_RESET) == 0)
		if( rto < get_timer(0))		// reset timeout ?
			return -1;

	outb(0xff, ioaddr + EN0_ISR);		// ack all intr
	return 0;
}

static int ax88796_probe( int ioaddr)
{
	// verify chip present
	if( inb(E8390_BASE) == 0xff)
		return -1;

	// Do a preliminary verification that we have a 8390 chipset
	outb(E8390_NODMA+E8390_PAGE1+E8390_STOP, ioaddr + E8390_CMD);	// set page 1 registers
	outb(0xff, ioaddr + EN0_COUNTER0);	// set 0xff at offset of EN0_COUNTER0(page0)
	outb(E8390_NODMA+E8390_PAGE0, ioaddr + E8390_CMD);	// switch to page 0
	inb(ioaddr + EN0_COUNTER0);	// clear the counter by reading
	if( inb(ioaddr + EN0_COUNTER0) != 0)
		return -1;

	return 0; 
}

int eth_init( bd_t *bd )
{
#ifdef DEBUG
	printf( "ax88796: eth_init\n");
#endif
	if( ax88796_probe(E8390_BASE) < 0)
	{
		printf("ax88796: ethernet chip not found\n");
		return 0;
	}
#ifdef DEBUG
	printf("ax88796: found at 0x%08x\n", E8390_BASE);
#endif
	if( ax88796_reset(E8390_BASE) < 0)
	{
		printf("ax88796: reset failed (no reset ack)\n");
		return 0;
	}

	// follow ax88796 init procedure
	outb(E8390_NODMA+E8390_PAGE0+E8390_STOP, E8390_BASE + E8390_CMD);	// switch to page 0
	outb( 0x01, E8390_BASE + EN0_DCFG);	// select word wide transfer
	outb( E8390_RXCONFIG, E8390_BASE + EN0_RXCR);   // irq high active, accept broadcasts
	
	// clear the remote byte count registers
	outb( 0x00, E8390_BASE + EN0_RCNTLO);
	outb( 0x00, E8390_BASE + EN0_RCNTHI);
        
	// set the transmit page and receive ring
	outb( 0x40, E8390_BASE + EN0_TPSR);		// tx start page register
	outb( 0x46, E8390_BASE + EN0_STARTPG);		// start page register
	outb( 0x80, E8390_BASE + EN0_STOPPG);		// stop page register
	outb( 0x46, E8390_BASE + EN0_BOUNDARY);
	
	// clear the pending interrupts and mask
	outb( 0xff, E8390_BASE + EN0_ISR);
	outb( ENISR_ALL, E8390_BASE + EN0_IMR);

	{
		int i;
		// set mac address
		outb(E8390_NODMA + E8390_PAGE1 + E8390_STOP, E8390_BASE + E8390_CMD);	// switch to page 1
	
		for(i = 0; i < ETHER_ADDR_LEN; i++)
		{
			outb( bd->bi_enetaddr[i], E8390_BASE + EN1_PHYS_SFT(i));
			if( inb( E8390_BASE + EN1_PHYS_SFT(i)) != bd->bi_enetaddr[i])
				printf( "ax88796: mac address read/write mismap %d\n", i);
		}
	}		
	outb( 0x47, E8390_BASE + EN1_CURPAG);		// set current page register
	outb(E8390_NODMA+E8390_PAGE0+E8390_START, E8390_BASE + E8390_CMD);      // switch to page 0 + start
	
	if( inb( E8390_BASE + EN0_GPI) & 0x02)		// set to full duplex mode
		outb( 0x80 | E8390_TXCONFIG, E8390_BASE + EN0_TXCR);
	else
		outb( E8390_TXCONFIG, E8390_BASE + EN0_TXCR);

	return 0;
}

// Send a data block via Ethernet
int eth_send(volatile void *packet, int length)
{
	unsigned txto;
#ifdef DEBUG
	printf( "ax88796: eth_send\n");
#endif
	outb(E8390_PAGE0+E8390_START+E8390_NODMA, E8390_BASE + E8390_CMD);	// switch to page 0
	outb(ENISR_RDC, E8390_BASE + EN0_ISR);		// set remote dma complete

	if( length & 0x01)	// round the length up for word writes
		length++;

	// setup remote dma transfer
	outb( length & 0xff, E8390_BASE + EN0_RCNTLO);
	outb( length >> 8, E8390_BASE + EN0_RCNTHI);
	outb( 0x00, E8390_BASE + EN0_RSARLO);
	outb( 0x40, E8390_BASE + EN0_RSARHI);

	//start data transfer
	outb( E8390_RWRITE+E8390_START, E8390_BASE + E8390_CMD);
	outsw( E8390_BASE + NE_DATAPORT, (unsigned short*)packet, length>>1);

	// write packet tx length
        outb( length & 0xff, E8390_BASE + EN0_TCNTLO);
	outb( length >> 8, E8390_BASE + EN0_TCNTHI);	

	// trigger tx
	outb( E8390_NODMA+E8390_TRANS+E8390_START, E8390_BASE + E8390_CMD);
	
	txto =  get_timer(0) + 2*CFG_HZ/10;	// tx timeout 200ms
	
	while( (inb( E8390_BASE + EN0_ISR) & (ENISR_RDC|ENISR_TX)) == 0)
	{
		if( txto < get_timer(0))
		{
			printf("ax88796: tramsmit timeout\n");
//			ax88796_reset(E8390_BASE);
			break;
		}
	}

	outb( ENISR_RDC, E8390_BASE + EN0_ISR);  // ack intr

	return 0;
}

static void ax88796_read( unsigned char* buffer, int count, unsigned ring_addr)
{
#ifdef DEBUG
	printf( "ax88796: reading %d bytes from ring at 0x%x\n", count, ring_addr);
#endif
	outb( E8390_NODMA+E8390_PAGE0+E8390_START, E8390_BASE + E8390_CMD);  // switch to page 0

	outb( count & 0xff, E8390_BASE + EN0_RCNTLO);	// write count of bytes
	outb( count >> 8, E8390_BASE + EN0_RCNTHI);
	outb( ring_addr & 0xff, E8390_BASE + EN0_RSARLO);	// set read start address
	outb( ring_addr >> 8, E8390_BASE + EN0_RSARHI);
	
	// read data from rx buffer
	outb( E8390_RREAD+E8390_START, E8390_BASE + E8390_CMD);
	insw( E8390_BASE + NE_DATAPORT, (unsigned short*)buffer, count>>1);

	if (count & 0x01)
		buffer[count-1] = inb( E8390_BASE + NE_DATAPORT);

	outb( ENISR_RDC, E8390_BASE + EN0_ISR);  // ack intr
}

// Get a data block via Ethernet
int eth_rx(void)
{
	int rxlen = 0;
	unsigned char curr_page, this_page;
	struct e8390_pkt_hdr rx_frame;
#ifdef DEBUG
	printf( "ax88796: eth_rx\n");
#endif
	// get the rx page (incoming packet pointer)
	outb(E8390_NODMA+E8390_PAGE1, E8390_BASE + E8390_CMD);	// switch to page 1, get curr page register
	curr_page = inb( E8390_BASE + EN1_CURPAG);
	outb(E8390_NODMA+E8390_PAGE0, E8390_BASE + E8390_CMD);	// switch to page 0
#ifdef DEBUG
	printf( "curr_page=0x%x\n", curr_page);
#endif	
	// remove one frame from the ring, boundary is always a page behind
	this_page = inb( E8390_BASE + EN0_BOUNDARY) + 1;
	if (this_page >= 0x80)
		this_page = 0x46;	// wrap around page register
#ifdef DEBUG
	printf( "this_page=0x%x\n", this_page);
#endif	
	if (this_page == curr_page)   // read all the frames
		return 0;

	// read page aligned packet header
	ax88796_read( (unsigned char*) &rx_frame, sizeof(struct e8390_pkt_hdr), this_page<<8);
#ifdef DEBUG
	printf( "8390hdr: status=0x%x, next=0x%x, count=0x%x\n", rx_frame.status, rx_frame.next, rx_frame.count);
#endif	
	rxlen = rx_frame.count - sizeof(struct e8390_pkt_hdr);

	// read packet data
	if(rxlen > (PKTSIZE_ALIGN+PKTALIGN))
		printf( "ax88796: bogus packet size %d\n", rxlen);
	
	if( (rx_frame.status & 0x0F) == ENRSR_RXOK)
		ax88796_read( (unsigned char*) NetRxPackets[0], rxlen, (this_page<<8) + sizeof(rx_frame));

	if( rx_frame.next > 0x80 && rx_frame.next <= 0x46)	// set boundary register
		rx_frame.next = 0x80;
		
	outb( rx_frame.next-1, E8390_BASE + EN0_BOUNDARY);
#ifdef DEBUG	
	printf( "setting boundary to 0x%x\n", inb( E8390_BASE + EN0_BOUNDARY));
#endif
	NetReceive(NetRxPackets[0], rxlen);

	return rxlen;
}

#endif	// COMMANDS & CFG_NET

#endif	// CONFIG_DRIVER_AX88796


