/*
 * Cirrus Logic CS8900A Ethernet
 *
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * Copyright (C) 1999 Ben Williamson <benw@pobox.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is loaded into SRAM in bootstrap mode, where it waits
 * for commands on UART1 to read and write memory, jump to code etc.
 * A design goal for this program is to be entirely independent of the
 * target board.  Anything with a CL-PS7111 or EP7211 should be able to run
 * this code in bootstrap mode.  All the board specifics can be handled on
 * the host.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "armboot.h"
#include "command.h"
#include "cs8900.h"
#include "net.h"

#ifdef CONFIG_DRIVER_CS8900

#if (CONFIG_COMMANDS & CFG_CMD_NET)


/* packet page register access functions */

#ifdef CS8900_BUS32
/* we don't need 16 bit initialisation on 32 bit bus */
#define get_reg_init_bus(b,x) get_reg(b,x)
#else
static unsigned short get_reg_init_bus(ulong base, int regno)
{
    /* force 16 bit busmode */
    volatile unsigned char c;
    c = CS8900_BUS16_0(base);
    c = CS8900_BUS16_1(base);
    c = CS8900_BUS16_0(base);
    c = CS8900_BUS16_1(base);
    c = CS8900_BUS16_0(base);

    CS8900_PPTR(base) = regno;
    return (unsigned short) CS8900_PDATA(base);
}
#endif

static unsigned short
get_reg(ulong base, int regno)
{
    CS8900_PPTR(base) = regno;
    return (unsigned short) CS8900_PDATA(base);
}


static void put_reg(ulong base, int regno, unsigned short val)
{
    CS8900_PPTR(base) = regno;
    CS8900_PDATA(base) = val;
}

static void eth_reset(ulong base)
{
    int tmo;
    unsigned short us;

    /* reset NIC */
    put_reg(base, PP_SelfCTL, get_reg(base, PP_SelfCTL) | PP_SelfCTL_Reset );

    /* wait for 30ms */
    udelay(30000);
    
    /* Wait until the chip is reset */
    tmo = get_timer(0) + 1 * CFG_HZ;
    while( (((us = get_reg_init_bus(base, PP_SelfSTAT)) & PP_SelfSTAT_InitD) == 0) && tmo < get_timer(0));
}

void cs8900_get_enetaddr(ulong base, uchar *addr)
{
    int i;
    /* verify chip id */
    if (get_reg_init_bus(base, PP_ChipID) != 0x630e)
	return;
    eth_reset(base); 
    if ((get_reg(base, PP_SelfST) & (PP_SelfSTAT_EEPROM | PP_SelfSTAT_EEPROM_OK)) == 
	      (PP_SelfSTAT_EEPROM|PP_SelfSTAT_EEPROM_OK)) {
	/* Load the MAC from EEPROM */      
	for (i=0; i<6/2; i++) {
	    unsigned int Addr;
	    Addr = get_reg(base, PP_IA+i*2);
	    addr[i*2] = Addr & 0xFF;
	    addr[i*2+1] = Addr >> 8;
	}
    }		
}

void cs8900_set_enetaddr( ulong base, uchar *addr)
{
	int i;
	/* verify chip id */
	if (get_reg_init_bus(base, PP_ChipID) != 0x630e)
		return;
	eth_reset(base);
	
	/* set the Ethernet address */
	for (i=0; i < 3; i++)
		put_reg( base, (PP_IA+i*2), addr[i*2] | (addr[i*2+1]<<8));
}

void eth_halt( void )
{
    /* disable transmitter/receiver mode */
    put_reg(CS8900_BASE, PP_LineCTL, 0);

    /* "shutdown" to show ChipID or kernel wouldn't find he cs8900 ... */
    get_reg_init_bus(CS8900_BASE, PP_ChipID);
}

int eth_init( bd_t *bd )
{
    
    /* verify chip id */
    if (get_reg_init_bus(CS8900_BASE, PP_ChipID) != 0x630e)
    {
	printf( "CS8900 Ethernet chip not found?!\n" );
	return 0;
    }

    /* reset & set the ethernet address */
    cs8900_set_enetaddr(CS8900_BASE, bd->bi_enetaddr);
	     
    /* receive only error free packets addressed to this card */
    put_reg(CS8900_BASE,PP_RxCTL, PP_RxCTL_IA | PP_RxCTL_Broadcast | PP_RxCTL_RxOK);

    /* do not generate any interrupts on receive operations */
    put_reg(CS8900_BASE,PP_RxCFG, 0);

    /* do not generate any interrupts on transmit operations */
    put_reg(CS8900_BASE,PP_TxCFG, 0);

    /* do not generate any interrupts on buffer operations */
    put_reg(CS8900_BASE,PP_BufCFG, 0);

    /* enable transmitter/receiver mode */
    put_reg(CS8900_BASE,PP_LineCTL, PP_LineCTL_Rx | PP_LineCTL_Tx);

    return 0;
}

/* Get a data block via Ethernet */
extern int eth_rx(void)
{
    int i;
    unsigned short rxlen;
    unsigned short *addr;
    unsigned short status;

    status = get_reg(CS8900_BASE,PP_RER);

    if ((status & PP_RER_RxOK) == 0)
      return 0;

    status = CS8900_RTDATA(CS8900_BASE);       /* stat */
    rxlen = CS8900_RTDATA(CS8900_BASE);        /* len */

    if(rxlen > PKTSIZE_ALIGN + PKTALIGN)
      printf("packet too big!\n");

    for(addr = (unsigned short *)NetRxPackets[0], i = rxlen >> 1; i > 0; i-- )
      *addr++ = CS8900_RTDATA(CS8900_BASE);
    if(rxlen & 1)
      *addr++ = CS8900_RTDATA(CS8900_BASE);

    /* Pass the packet up to the protocol layers. */
    NetReceive(NetRxPackets[0], rxlen);

    return rxlen;
}

/* Send a data block via Ethernet. */
extern int eth_send(volatile void *packet, int length)
{
    volatile unsigned short *addr;
    int tmo;
    unsigned short s;

retry:
    /* initiate a transmit sequence */
    CS8900_TxCMD(CS8900_BASE) = PP_TxCmd_TxStart_Full;
    CS8900_TxLEN(CS8900_BASE) = length;
    
    /* Test to see if the chip has allocated memory for the packet */
    if ((get_reg(CS8900_BASE,PP_BusSTAT) & PP_BusSTAT_TxRDY) == 0) {    
	/* Oops... this should not happen! */
	printf("cs: unable to send packet; retrying...\n");
	for (tmo = get_timer(0) + 5 * CFG_HZ; get_timer(0) < tmo; )
	  /*NOP*/;
	eth_reset(CS8900_BASE);
	goto retry; 
    }

    /* Write the contents of the packet */
    /* assume even number of bytes */
    for(addr = packet; length > 0; length -= 2 )
      CS8900_RTDATA(CS8900_BASE) = *addr++;

    /* wait for transfer to succeed */
    tmo = get_timer(0) + 5 * CFG_HZ;
    while((s = get_reg(CS8900_BASE,PP_TER) & ~0x1F) == 0) 
    {
	if (get_timer(0) >= tmo)
	  break;
    }
     
      /* nothing */ ;
    if ((s & (PP_TER_CRS | PP_TER_TxOK)) != PP_TER_TxOK) {
	printf("\ntransmission error %#x\n", s);
    }

    return 0;
}

#endif /* COMMANDS & CFG_NET */

#endif /* CONFIG_DRIVER_CS8900 */
